// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_fw_offensive_kick.h"

#include "body_kick_to_corner.h"

#include "attributeparam.h"

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_point.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>

#include <cmath>

//#define DEBUG2

/*-------------------------------------------------------------------*/
/*!
  
*/

// Shoot -> Dribble -> Pass の順番で行動を決める
bool
Bhv_FWOffensiveKick::execute( rcsc::PlayerAgent * agent )
{
  
  rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_BasicOffensiveKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
    const rcsc::PlayerObject * nearest_opp
        = ( opps.empty()
            ? static_cast< rcsc::PlayerObject * >( 0 )
            : opps.front() );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
                                             ? nearest_opp->pos()
                                             : rcsc::Vector2D( -1000.0, 0.0 ) );
    //相手キーパー
    const rcsc::PlayerObject *opg = wm.getOpponentGoalie();

    rcsc::Vector2D pass_point;
    rcsc::Vector2D shoot_point;

    // 特性パラメータ取得
    typedef boost::shared_ptr<AttributeParam> ParamPtr ;
    ParamPtr param = ParamPtr( new AttributeParam() ) ;
    
    double s_param = param->get_shoot_param( wm.time().cycle() ) ;
    double p_param = param->get_pass_param( wm.time().cycle() ) ;
    double d_param = param->get_dribble_param( wm.time().cycle() ) ;

    // 以前は「ペナルティエリア内」だったが，ベクトル角が10度以上だったらに
    // 変える．
    if( opg ) {
      const rcsc::Vector2D self2gk = opg->pos() - wm.self().pos() ;
      
      //ゴールポストへのベクトル（実際よりかなり広くしてある）
      const rcsc::Vector2D goal_l( rcsc::ServerParam::i().pitchHalfLength(), 
				   10.0 ) ;
      const rcsc::Vector2D goal_r( rcsc::ServerParam::i().pitchHalfLength(), 
				   -10.0 ) ;
      const rcsc::Vector2D self2gl = goal_l - wm.self().pos() ;
      const rcsc::Vector2D self2gr = goal_r - wm.self().pos() ;
      
      /*
      double naisekil = self2gl.innerProduct( self2gk )  ;
      double naisekir = self2gr.innerProduct( self2gk )  ;
      
      double vec_length_l = self2gl.r() * self2gk.r() ;
      double vec_length_r = self2gr.r() * self2gk.r() ;
      
      //      double gl_cosval = naisekil / vec_length_l ;
      //      double gr_cosval = naisekir / vec_length_r ;
      */

      const double gl_cosval = self2gl.innerProduct( self2gk ) /
	( self2gl.r() * self2gk.r() ) ;
      const double gr_cosval = self2gr.innerProduct( self2gk ) /
	( self2gr.r() * self2gk.r() ) ;
      
      const double self2gl_vecdeg = rcsc::AngleDeg::acos_deg( gl_cosval ) ;
      const double self2gr_vecdeg = rcsc::AngleDeg::acos_deg( gr_cosval ) ;

#ifdef DEBUG2	 
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d:Debug: vecDeg_l : %lf : %lf"
			  ,__FILE__, __LINE__, self2gl_vecdeg, gl_cosval ) ;
      
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d:Debug: gr_cosval : %lf"
			  ,__FILE__, __LINE__, gr_cosval  ) ;
#endif
      
      double shootp_flag = 1.0 ;
      
      if ( wm.self().pos().x > 30.0 ){
	if ( std::max( self2gl_vecdeg, self2gl_vecdeg ) > s_param ) {
	  
	  double shoot_power = 10.0 ;
	  //	  wm.self().pos().x > 30.0 ? shoot_power = 1
	  //	    : shoot_power = 10.0 ; // default 7.0
	  
	  // シュートコースを決定する でもそんな大層なものでもない
	  self2gl_vecdeg >= self2gr_vecdeg ? shootp_flag *= 1.0 
	    : shootp_flag *= -1.0 ;
	  
	  const double shootp_y = 5.0 * shootp_flag ;
	  
	  rcsc::dlog.addText( rcsc::Logger::TEAM,
			      "%s:%d:Debug: Shoot_p : %lf"
			      ,__FILE__, __LINE__, shootp_y  ) ;
	  
	  shoot_point.assign( 52.5, shootp_y ) ;
	  
	  if ( rcsc::Body_KickMultiStep( shoot_point, 3.0 ).execute( agent ) ){
	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"SHOOT(visible GK)!: %lf : %lf : %lf "
				,shoot_point.y, opg->pos().y, shoot_power ) ;
	    agent->setNeckAction( new rcsc::Neck_TurnToPoint( shoot_point ) );
	    return true ;
	  }
	}
      }
    }
    
    //キーパーが見えない && PA内ではゴールに向かってシュート
    if( wm.self().pos().x > 36.0 && wm.self().pos().y < 30.0 
	&& wm.self().pos().y > -30.0 ) {  // default 22.0
      
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: On Penalty Area : %lf %lf"
			  ,__FILE__, __LINE__,
			  wm.self().pos().x, wm.self().pos().y ) ;
      agent->setNeckAction( new rcsc::Neck_TurnToPoint( shoot_point ) );

      //      double y_pos = wm.self().pos().y ;
      
      if( wm.self().pos().y > 20.0 || wm.self().pos().y < -20.0 ) {  //default 15.0
	shoot_point.assign( 40.0, 0.0 ) ;
	if ( rcsc::Body_KickMultiStep( shoot_point, 3.0 ).execute( agent ) ) {
	  rcsc::dlog.addText( rcsc::Logger::TEAM,
			      "%s:%d: Cross: %lf"
			      ,__FILE__, __LINE__ ) ;
	  agent->setNeckAction( new rcsc::Neck_TurnToPoint( shoot_point ) );
	  return true ;
	}
      }
      
      /*
      const rcsc::AngleDeg shoot_angle1 = ( opg->pos() - wm.self().pos() ).th();
      
      rcsc::dlog.addText( rcsc::Logger::TEAM,
			  "%s:%d: Shoot Angle: %lf"
			  ,__FILE__, __LINE__, shoot_angle1.abs() ) ;
      
      agent->debugClient().addMessage( "Shoot Angle: %lf", shoot_angle1.abs() );
      
      if ( shoot_angle1.abs() > s_param ) {
	rcsc::dlog.addText( rcsc::Logger::TEAM,
			    "%s:%d: Shoot Activate: %lf"
			    ,__FILE__, __LINE__, shoot_angle1.abs() ) ;
      */
      shoot_point.assign( 52.5, 0.0 ) ;
      
      if ( rcsc::Body_KickMultiStep( shoot_point, 3.0 ).execute( agent ) ) {
	rcsc::dlog.addText( rcsc::Logger::TEAM,
			    "SHOOT(unvisible GK)!: %lf"
			    , shoot_point.y ) ;
	agent->setNeckAction( new rcsc::Neck_TurnToPoint( shoot_point ) );
	return true ;
      }
    } 

	
	//if ( nearest_opp_dist < 3.0 ) //default 5.0
	//&& nearest_opp_dist > ( rcsc::ServerParam::i().tackleDist()
	//+ rcsc::ServerParam::i().defaultPlayerSpeedMax() * 1.5 )
	// && wm.self().body().abs() < 90.0 ) //default 70.0

	// dribble to my body dir
	if ( nearest_opp_dist < d_param ) //default 5.0
	  {

	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"%s:%d: Dribble Activate(1)"
				,__FILE__, __LINE__ ) ;
	    
	    const rcsc::Vector2D body_dir_drib_target
	      = wm.self().pos()
	      + rcsc::Vector2D::polar2vector(5.0, wm.self().body());
	    
	    int max_dir_count = 0;
	    wm.dirRangeCount( wm.self().body(), 20.0, &max_dir_count, NULL, NULL );
	    
	    if ( body_dir_drib_target.x < rcsc::ServerParam::i().pitchHalfLength() - 1.0
		 && body_dir_drib_target.absY() < rcsc::ServerParam::i().pitchHalfWidth() - 1.0
		 && max_dir_count < 3
		 )
	      {
		// check opponents
		// 5m, +-10 degree (default 10m, +-30 degree)
		const rcsc::Sector2D sector( wm.self().pos(),
					     0.5, 10.0,  //default 0.5, 10.0
					     wm.self().body() - 10.0,
					     wm.self().body() + 10.0 );
		// opponent check with goalie
		if ( ! wm.existOpponentIn( sector, 10, true ) )
		  {
		    rcsc::dlog.addText( rcsc::Logger::TEAM,
					"%s:%d: dribble to my body dir"
					,__FILE__, __LINE__ );
		    agent->debugClient().addMessage( "OffKickDrib(1)" );
		    rcsc::Body_Dribble( body_dir_drib_target,
					1.0,
					rcsc::ServerParam::i().maxPower(),
					2
					).execute( agent );
		    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
		    return true;
		  }
	      }
	  }
	
	rcsc::Vector2D drib_target( 50.0, wm.self().pos().absY() );
	if ( drib_target.y < 20.0 ) drib_target.y = 20.0;
	if ( drib_target.y > 29.0 ) drib_target.y = 27.0;
	if ( wm.self().pos().y < 0.0 ) drib_target.y *= -1.0;
	const rcsc::AngleDeg drib_angle = ( drib_target - wm.self().pos() ).th();
	
	// opponent is behind of me
	if ( nearest_opp_pos.x < wm.self().pos().x + 1.0 )
	  {
	    // check opponents
	    // 15m, +-10 degree (default +-30 degree)
	    const rcsc::Sector2D sector( wm.self().pos(),
					 0.5, 15.0,
					 drib_angle - 10.0,
					 drib_angle + 10.0 );
	    // opponent check with goalie
	    if ( ! wm.existOpponentIn( sector, 10, true ) )
	      {
		const int max_dash_step
		  = wm.self().playerType()
		  .cyclesToReachDistance( wm.self().pos().dist( drib_target ) );
		if ( wm.self().pos().x > 35.0 )
		  {
		    drib_target.y *= ( 10.0 / drib_target.absY() );
		  }
		
		rcsc::dlog.addText( rcsc::Logger::TEAM,
				    "%s:%d: fast dribble to (%.1f, %.1f) max_step=%d"
				    ,__FILE__, __LINE__,
				    drib_target.x, drib_target.y,
				    max_dash_step );
		agent->debugClient().addMessage( "OffKickDrib(2)" );
		rcsc::Body_Dribble( drib_target,
				    1.0,
				    rcsc::ServerParam::i().maxPower(),
				    std::min( 5, max_dash_step )
				    ).execute( agent );
	      }
	    else
	      {
		rcsc::dlog.addText( rcsc::Logger::TEAM,
				    "%s:%d: slow dribble to (%f, %f)"
				    ,__FILE__, __LINE__,
				    drib_target.x, drib_target.y );
		agent->debugClient().addMessage( "OffKickDrib(3)" );
		rcsc::Body_Dribble( drib_target,
				    1.0,
				    rcsc::ServerParam::i().maxPower(),
				    2
				    ).execute( agent );
		
	      }
	    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	    return true;
	  }
	
	// opp is far from me
	if ( nearest_opp_dist > d_param ) // default 5.0
	  {
	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"%s:%d: Dribble Activate(2) opp far. dribble(%f, %f), "
				,__FILE__, __LINE__,
				drib_target.x, drib_target.y );
	    agent->debugClient().addMessage( "OffKickDrib(4)" );
	    rcsc::Body_Dribble( drib_target,
				1.0,
				rcsc::ServerParam::i().maxPower(), //* 0.9,
				1
				).execute( agent );
	    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	    return true;
	  }
	
	// opp is far from me
	
	if ( nearest_opp_dist > std::max( ( d_param / 2.0 ), 1.0 ) ) // default 3.0
	  {
	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"%s:%d: Dribble Activate(3) opp far. dribble(%f, %f)"
				,__FILE__, __LINE__,
				drib_target.x, drib_target.y );
	    agent->debugClient().addMessage( "OffKickDrib(5)" );
	    rcsc::Body_Dribble( drib_target,
				1.0,
				rcsc::ServerParam::i().maxPower() * 0.8, //* 0.2,
				1
				).execute( agent );
	    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	    return true;
	  }
	
	if ( nearest_opp_dist > std::max( ( d_param - 4.5 ), 0.5) ) //default 1.5 and tkzwtks default 0.5
	  {
	    rcsc::dlog.addText( rcsc::Logger::TEAM,
				"%s:%d: Hold Activate"
				,__FILE__, __LINE__ );
	    agent->debugClient().addMessage( "OffKickHold" );
	    rcsc::Body_HoldBall().execute( agent );
	    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	    return true;
	  }
	
	// "get_best_pass"でパスコースを探し，
	// そのパスポイントが前方にある場合に安全かどうか確かめる
	if ( rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL ) )
	  {
	    if ( pass_point.x > wm.self().pos().x - 1.0 )
	      {
		bool safety = true;
		const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
		for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
		      it != opps_end;
		      ++it )
		  {
		    if ( (*it)->pos().dist( pass_point ) < p_param )
		      // Default dist( pass_point ) < 4.0
		      {
			safety = false;
		      }
		  }
		
		if ( safety )
		  {
		    rcsc::dlog.addText( rcsc::Logger::TEAM,
					"%s:%d: SafetyPass Activate,(do best pass)"
					,__FILE__, __LINE__ );
		    agent->debugClient().addMessage( "OffKickPass(1)" );
		    rcsc::Body_Pass().execute( agent );
		    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
		    return true;
		  }
	      }
	  }

    
    // ここから共通．
    // 相手がかなり近い場合，パスができればパスを試みる
    // can pass
    if ( rcsc::Body_Pass().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "OffKickPass(3)" );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return true;
    }

if ( wm.self().pos().x > wm.offsideLineX() - 10.0 )
  {
    rcsc::dlog.addText( rcsc::Logger::TEAM,
			"%s:%d: kick near side"
			,__FILE__, __LINE__ );
    agent->debugClient().addMessage( "OffKickToCorner" );
    Body_KickToCorner( (wm.self().pos().y < 0.0) ).execute( agent );
    agent->setNeckAction( new rcsc::Neck_ScanField() );
	  }
 else
   {
     rcsc::dlog.addText( rcsc::Logger::TEAM,
			 "%s:%d: advance"
			 ,__FILE__, __LINE__ );
     agent->debugClient().addMessage( "OffKickAdvance" );
     rcsc::Body_AdvanceBall().execute( agent );
     agent->setNeckAction( new rcsc::Neck_ScanField() );
   }

return true;
}

